#!/usr/bin/perl -T
#
#   @(#)myphat.cgi	1.20 06/01/17
#
#   Perl CGI script for browsing the PhatNoise(R) Audio System DMS
#   music collection, based on the *.db file from the DMS PHTDTA
#   partition.
#
#   Copyright (C) 2006  Ti Kan
#   http://www.amb.org/ti
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

use strict;
use warnings;
use CGI qw/:all unescape/;
# use CGI::Carp qw(fatalsToBrowser);	# Uncomment this when debugging

# Local customizations begin here
# ----------------------------------------------------

# My name
my $my_name = "Jon Doe";

# URL to the logo image
my $logo_url = "http://www.yourdomain.com/musiclogo.png";

# URL to the downward pointing arrow icon image
my $downarrow_url = "http://www.yourdomain.com/downarrow.gif";

# Path to the directory where the *.db files are kept
my $db_dir = "./phat_dbdir";

# Gracenote CDDB search URL
my $cddb_search_url = "http://www.gracenote.com/music/search.html";

# Font size and style for table area text
my $font_size = '-2';
my $font_face = 'arial,helvetica';

# Colors
my $text_color  = '#000000';	# Text color
my $link_color  = '#0000FF';	# Hyperlink color
my $sur_bgcolor = '#8899DD';	# Surrounding background color
my $tbl_bgcolor = '#FEFEDD';	# General table area background color
my $frm_bgcolor = '#DFDFDF';	# Frame areas background color
my $sr1_bgcolor = '#FFFFCC';	# Active sort mode heading background color
my $sr2_bgcolor = '#DFDFDF';	# Inactive sort mode heading background color
my $tt1_bgcolor = '#BBFFDD';	# Track list background color 1
my $tt2_bgcolor = '#AADDEE';	# Track list background color 2

# Name of top level frame.  This is important when this page is to be
# displayed under a frames environment.
my $top_frame = '_top';

# ----------------------------------------------------
# End local customizations


# URL to myself (this script)
my $my_url = url();

# Passed variables
my $genre = urlDecode(param("genre") || '');
my $artist = urlDecode(param("artist") || '');
my $album = urlDecode(param("album") || '');
my $submit = param("submit") || '';
my $subframe = param("subframe") || '';
my $sort = param("sort") || '';

# URL-encoded versions of genre, artist and album names
my $e_genre = urlEncode($genre);
my $e_artist = urlEncode($artist);
my $e_album = urlEncode($album);

# DMS database hash tables
my %genres = ();
my %artists = ();
my %albums = ();
my %tracks = ();

# Counts
my $ngenres = 0;
my $nartists = 0;
my $nalbums = 0;
my $ntracks = 0;

my $show_all = 0;
if ($genre eq '' && $artist eq '' && $album eq '') {
    # Default display mode is to show everything
    $show_all = 1;
}

if ($sort eq '') {
    # Default sort mode is "album"
    $sort = 'album';
}

# Enable autoflush
$| = 1;

# Start HTML
&startPage;

# Top title
if ($subframe eq '') {
    &setUpGenres;
    &setUpArtists;
    &setUpAlbums;
    &setUpTracks;

    print table({ -cellpadding => "6", -cellspacing => "0", -border => "0" },
		Tr(td({ -align => "left", -valign => "middle" },
		      img({ -src => $logo_url, -border => "0", -alt => "" })),
		   td({ -align => "left", -valign => "top" },
		      font({ -face => $font_face },
			   b("${my_name}'s Music Collection")),
		      font({ -size => $font_size, -face => $font_face },
			   br(),
			   "Select a genre, artist or album from ",
			   "the lists below."), "\n",
		      br(), "\n"))),
	  "<div align=\"center\">\n",
	  "<table cellpadding=\"2\" cellspacing=\"0\" ",
	  "border=\"0\" width=\"100%\" bgcolor=\"${tbl_bgcolor}\">\n",
	  "<tr>\n";
}

# Genre list
if ($subframe eq '') {
    print "<td width=\"33%\">\n",
	  "<table cellpadding=\"2\" cellspacing=\"0\" border=\"0\">\n",
	  "<tr>\n",
	  "<th align=\"left\">\n",
	  font({ -size => $font_size, -face => $font_face },
	       "Genres ($ngenres)"),
	  "</th>\n",
	  "</tr>\n",
	  "<tr>\n",
	  "<td align=\"left\">\n",
	  "<iframe name=\"genre\" src=\"${my_url}?",
	  "submit=${submit}&sort=${sort}&subframe=genre&",
	  "genre=${e_genre}&artist=${e_artist}&album=${e_album}\" ",
	  "frameborder=\"1\" align=\"middle\" ",
	  "scrolling=\"yes\" marginwidth=\"1\" marginheight=\"1\" ",
	  "width=\"100%\" height=\"100%\">\n";
    &setUpGenres;
    &displayGenres;
    print "</iframe>\n",
	  "</td>\n</tr>\n</table>\n</td>\n";
}
elsif ($subframe eq 'genre') {
    &setUpGenres;
    &displayGenres;
}

# Artist list
if ($subframe eq '') {
    print "<td width=\"33%\">\n",
	  "<table cellpadding=\"2\" cellspacing=\"0\" border=\"0\">\n",
	  "<tr>\n",
	  "<th align=\"left\">\n",
	  font({ -size => $font_size, -face => $font_face },
	       "Artists ($nartists)"),
	  "</th>\n",
	  "</tr>\n",
	  "<tr>\n",
	  "<td align=\"left\">\n",
	  "<iframe name=\"artist\" src=\"${my_url}?",
	  "submit=${submit}&sort=${sort}&subframe=artist&",
	  "genre=${e_genre}&artist=${e_artist}&album=${e_album}\" ",
	  "frameborder=\"1\" align=\"middle\" ",
	  "scrolling=\"yes\" marginwidth=\"1\" marginheight=\"1\" ",
	  "width=\"100%\" height=\"100%\">\n";
    &setUpArtists;
    &displayArtists;
    print "</iframe>\n",
	  "</td>\n</tr>\n</table>\n</td>\n";
}
elsif ($subframe eq 'artist') {
    &setUpArtists;
    &displayArtists;
}

# Album list
if ($subframe eq '') {
    print "<td width=\"33%\">\n",
	  "<table cellpadding=\"2\" cellspacing=\"0\" border=\"0\">\n",
	  "<tr>\n",
	  "<th align=\"left\">\n",
	  font({ -size => $font_size, -face => $font_face },
	       "Albums ($nalbums)"),
	  "</th>\n",
	  "</tr>\n",
	  "<tr>\n",
	  "<td align=\"left\">\n",
	  "<iframe name=\"album\" src=\"${my_url}?",
	  "submit=${submit}&sort=${sort}&subframe=album&",
	  "genre=${e_genre}&artist=${e_artist}&album=${e_album}\" ",
	  "frameborder=\"1\" align=\"middle\" ",
	  "scrolling=\"yes\" marginwidth=\"1\" marginheight=\"1\" ",
	  "width=\"100%\" height=\"100%\">\n";
    &setUpAlbums;
    &displayAlbums;
    print "</iframe>\n",
	  "</td>\n</tr>\n</table>\n</td>\n";
}
elsif ($subframe eq 'album') {
    &setUpAlbums;
    &displayAlbums;
}

# Track list
if ($subframe eq '') {
    print "</tr>\n",
	  "<tr>\n",
	  "<th colspan=\"3\" align=\"left\" width=\"100%\">\n",
	  hr(),
	  font({ -size => $font_size, -face => $font_face },
	       "Tracks ($ntracks)"),
	  "</th>\n",
	  "</tr>\n",
	  "<tr>\n",
	  "<td colspan=\"3\" align=\"center\" ",
	  "valign=\"top\" width=\"100%\">\n";
	  &displayTracks;
    print "</td>\n</tr>\n</table>\n</div>\n";

    print div({ -align => "left" },
	      hr(),
	      font({ -size => $font_size, -face => $font_face },
		   "Web interface Copyright &copy; 2006 ",
		   a({ -href => "http://www.amb.org/ti/",
		       -target => "_blank" },
		     "Ti Kan"), ", All rights reserved. ",
		   "@(#)myphat.cgi	1.20 06/01/17", br(), "\n",
		   a({ -href => "http://www.gracenote.com/",
		       -target => "_blank" },
		     "Gracenote CDDB&reg; Music Recognition Service"),
		   br(), "\n",
		   a({ -href => "http://www.amb.org/xmcd/",
		       -target => "_blank" },
		     "Xmcd CD player/ripper software for Linux and Unix"),
		   br(), "\n",
		   a({ -href => "http://www.phatnoise.com/",
		       -target => "_blank" },
		     "PhatNoise&reg; Car Audio System"), br(), "\n")), "\n";
}

# End HTML
&endPage;


# Functions

sub startPage
{
    print header(-type => 'text/html', -expires => 'now');

    if ($subframe eq '') {
	print start_html(
	      # Page title
	      -title => "${my_name}'s Music Collection",
	      # Colors
	      -text => $text_color,
	      -link => $link_color,
	      -vlink => $link_color,
	      -alink => $link_color,
	      -bgcolor => $sur_bgcolor,
	      -head => [
		  meta({ -http_equiv => 'EXPIRES', -content => '0' }),
		  Link({ -rel => 'SHORTCUT ICON',
			 -href => 'http://www.amb.org/ti/favicon.ico'})
	      ],
	), "\n";
    }
    else {
	print start_html(
	      # Page title
	      -title => "$subframe",
	      # Colors
	      -text => $text_color,
	      -link => $link_color,
	      -vlink => $link_color,
	      -alink => $link_color,
	      -bgcolor => $frm_bgcolor,
	      -head => [
		  meta({ -http_equiv => 'EXPIRES', -content => '0' })
	      ],
	), "\n";
    }
}


sub endPage
{
    print end_html(), "\n";
}


sub setUpGenres
{
    my @fields;
    my $key;
    my $line = 0;
    my $q_genre;
    my $q_artist;
    my $q_album;

    if (open(GENRE, "${db_dir}/genre.db")) {
	while (<GENRE>) {
	    $line++;
	    if (/^#/ || /^$[ 	]*/) {
		next; 	# comment or blank line
	    }
	    chomp;
	    @fields = split /\t/;
	    my $nfields = @fields;
	    if ($nfields != 9) {
		# Invalid number of fields
		die "Error in genre.db: invalid number of fields, line $line";
	    }

	    $q_genre = stripQuestionMark($fields[4]);
	    $q_artist = stripQuestionMark($fields[2]);
	    $q_album = stripQuestionMark($fields[3]);

	    if ($show_all == 1 ||
		($genre ne '' && $genre eq $q_genre) ||
		($artist ne '' && $artist eq $q_artist) ||
		($album ne '' && $album eq $q_album)) {
		$genres{$q_genre} = urlEncode($q_genre);
	    }
	}
	close GENRE;

	$ngenres = (keys %genres);
    }
    else {
	die("Fatal error: Cannot open genre.db!");
    }
}


sub setUpArtists
{
    my @fields;
    my $key;
    my $line = 0;
    my $q_genre;
    my $q_artist;
    my $q_album;

    if (open(ARTIST, "${db_dir}/artist.db")) {
	while (<ARTIST>) {
	    $line++;
	    if (/^#/ || /^$[ 	]*/) {
		next; 	# comment or blank line
	    }
	    chomp;
	    @fields = split /\t/;
	    my $nfields = @fields;
	    if ($nfields != 9) {
		# Invalid number of fields
		die "Error in artist.db: invalid number of fields, line $line";
	    }

	    $q_genre = stripQuestionMark($fields[4]);
	    $q_artist = stripQuestionMark($fields[2]);
	    $q_album = stripQuestionMark($fields[3]);

	    if ($show_all == 1 ||
		($genre ne '' && $genre eq $q_genre) ||
		($artist ne '' && $artist eq $q_artist) ||
		($album ne '' && $album eq $q_album)) {
		$artists{$q_artist} = urlEncode($q_artist);
	    }
	}
	close ARTIST;

	$nartists = (keys %artists);
    }
    else {
	die("Fatal error: Cannot open artist.db!");
    }
}


sub setUpAlbums
{
    my @fields;
    my $key;
    my $line = 0;
    my $q_genre;
    my $q_artist;
    my $q_album;

    if (open(ALBUM, "${db_dir}/album.db")) {
	while (<ALBUM>) {
	    $line++;
	    if (/^#/ || /^$[ 	]*/) {
		next; 	# comment or blank line
	    }
	    chomp;
	    @fields = split /\t/;
	    my $nfields = @fields;
	    if ($nfields != 9) {
		# Invalid number of fields
		die "Error in album.db: invalid number of fields, line $line";
	    }

	    $q_genre = stripQuestionMark($fields[4]);
	    $q_artist = stripQuestionMark($fields[2]);
	    $q_album = stripQuestionMark($fields[3]);

	    if ($show_all == 1 ||
		($genre ne '' && $genre eq $q_genre) ||
		($artist ne '' && $artist eq $q_artist) ||
		($album ne '' && $album eq $q_album)) {
		$albums{$q_album} = urlEncode($q_album);
	    }
	}
	close ALBUM;

	$nalbums = (keys %albums);
    }
    else {
	die("Fatal error: Cannot open album.db!");
    }
}


sub setUpTracks
{
    my @fields;
    my $key;
    my $line = 0;
    my $q_genre;
    my $q_artist;
    my $q_album;
    my $q_title;
    my $sortstr;

    #
    # Use album.db instead of tracks.db because it is likely to
    # be better sorted
    #
    if (open(ALBUM, "${db_dir}/album.db")) {
	while (<ALBUM>) {
	    $line++;
	    if (/^#/ || /^$[ 	]*/) {
		next; 	# comment or blank line
	    }
	    chomp;
	    @fields = split /\t/;
	    my $nfields = @fields;
	    if ($nfields != 9) {
		# Invalid number of fields
		die "Error in album.db: invalid number of fields, line $line";
	    }

	    $q_genre = stripQuestionMark($fields[4]);
	    $q_artist = stripQuestionMark($fields[2]);
	    $q_album = stripQuestionMark($fields[3]);
	    $q_title = stripQuestionMark($fields[1]);

	    foreach $key (sort keys %albums) {
		if ($key eq $q_album) {
		    if ($sort eq 'genre') {
			$sortstr = $q_genre .  " - " .
				   $q_album .  " - " .
				   $fields[7] .  " - " .
				   $q_artist .  " - " .
				   $q_title;
		    }
		    elsif ($sort eq 'artist') {
			$sortstr = $q_artist .  " - " .
				   $fields[7] .  " - " .
				   $q_title .  " - " .
				   $q_album .  " - " .
				   $q_genre;
		    }
		    elsif ($sort eq 'title') {
			$sortstr = $q_title .  " - " .
				   $q_album . '-' .
				   $fields[7] .  " - " .
				   $q_artist .  " - " .
				   $q_genre;
		    }
		    elsif ($sort eq 'track') {
			$sortstr = $fields[7] .  " - " .
				   $q_title .  " - " .
				   $q_album . '-' .
				   $q_artist .  " - " .
				   $q_genre;
		    }
		    else {
			$sortstr = $q_album .  " - " .
				   $fields[7] .  " - " .
				   $q_artist .  " - " .
				   $q_title . '-' .
				   $q_genre;
		    }

		    # The whole db line is stored
		    $tracks{$sortstr} = $_;
		}
	    }
	}
	close ALBUM;

	$ntracks = (keys %tracks);
    }
    else {
	die("Fatal error: Cannot open album.db!");
    }
}


sub displayGenres
{
    my $key = '';
    my $go_url = '';

    print a({ -href => "${my_url}?sort=${sort}", -target => ${top_frame},
	      -style => "text-decoration:none" },
	    font({ -size => $font_size, -face => $font_face },
		 nobr('(Click here for all genres)'))), br(), "\n";

    $go_url = "${my_url}?submit=genre&sort=${sort}&genre=";
    foreach $key (sort keys %genres) {
	print a({ -href => "${go_url}$genres{$key}", -target => ${top_frame},
		  -style => "text-decoration:none" },
		font({ -size => $font_size, -face => $font_face },
		     nobr(escapeHTML($key)))), br(), "\n";
    }
}


sub displayArtists
{
    my $key = '';
    my $go_url = '';

    print a({ -href => "${my_url}?sort=${sort}", -target => ${top_frame},
	      -style => "text-decoration:none" },
	    font({ -size => $font_size, -face => $font_face },
		 nobr('(Click here for all artists)'))), br(), "\n";

    $go_url = "${my_url}?submit=genre&sort=${sort}&artist=";
    foreach $key (sort keys %artists) {
	print a({ -href => "${go_url}$artists{$key}", -target => ${top_frame},
		  -style => "text-decoration:none" },
		font({ -size => $font_size, -face => $font_face },
		     nobr(escapeHTML($key)))), br(), "\n";
    }
}


sub displayAlbums
{
    my $key = '';
    my $go_url = '';

    print a({ -href => "${my_url}?sort=${sort}", -target => ${top_frame},
	      -style => "text-decoration:none" },
	    font({ -size => $font_size, -face => $font_face },
		 nobr('(Click here for all albums)'))), br(), "\n";

    $go_url = "${my_url}?submit=genre&sort=${sort}&album=";
    foreach $key (sort keys %albums) {
	print a({ -href => "${go_url}$albums{$key}", -target => ${top_frame},
		  -style => "text-decoration:none" },
		font({ -size => $font_size, -face => $font_face },
		     nobr(escapeHTML($key)))), br(), "\n";
    }
}


sub displayTracks
{
    my @fields;
    my $key;
    my $bgcolor = '';
    my $curr_album;
    my $u_genre = '-';
    my $u_artist = '-';
    my $u_album = '-';
    my $u_title = '-';
    my $q_genre;
    my $q_artist;
    my $q_album;
    my $q_title;
    my $prev_genre = '';
    my $prev_artist = '';
    my $prev_album = '';
    my $urlbase = '';

    $urlbase = "${my_url}?submit=$submit&";
    $urlbase .= "genre=${e_genre}&artist=${e_artist}&album=${e_album}";

    # Track list table
    print "<table cellpadding=\"2\" cellspacing=\"0\" border=\"1\" ",
	  "width=\"100%\">\n",
	  "<tr>\n";

    foreach $key ("Artist", "Album", "Genre", "Track", "Title") {
	if ($sort eq lc $key) {
	    print th({ -align => "center", -valign => "top",
		       -bgcolor => ${sr1_bgcolor} },
		     font({ -size => $font_size, -face => $font_face },
			  nobr("${key}&nbsp;",
			       img({ -src => $downarrow_url,
				     -border => "0", -alt => "0" })))),
		  "\n";
	}
	else {
	    print th({ -align => "center", -valign => "top",
		       -bgcolor => ${sr2_bgcolor} },
		     a({ -href => "${urlbase}&sort=artist",
			 -target => ${top_frame},
			 -style => "text-decoration:none" },
		       font({ -size => $font_size, -face => $font_face },
			    nobr("${key}&nbsp;",
				 img({ -src => $downarrow_url,
				       -border => "0", -alt => "0" }))))),
		  "\n";
	}
    }

    print "</tr>\n";

    $curr_album = '';

    foreach $key (sort keys %tracks) {
	@fields = split /\t/, $tracks{$key};

	$q_genre = stripQuestionMark($fields[4]);
	$q_artist = stripQuestionMark($fields[2]);
	$q_album = stripQuestionMark($fields[3]);
	$q_title = stripQuestionMark($fields[1]);

	$u_genre = '';
	if ($prev_genre ne $q_genre) {
	    $prev_genre = $q_genre;
	    $u_genre = urlEncode($q_genre);
	}
	$u_artist = '';
	if ($prev_artist ne $q_artist) {
	    $prev_artist = $q_artist;
	    $u_artist = urlEncode($q_artist);
	}
	$u_album = '';
	if ($prev_album ne $q_album) {
	    $prev_album = $q_album;
	    $u_album = urlEncode($q_album);
	}
	$u_title = urlEncode($q_title);

	if ($q_album ne $curr_album) {
	    $curr_album = $q_album;
	    if ($bgcolor eq $tt2_bgcolor) {
		$bgcolor = $tt1_bgcolor;
	    }
	    else {
		$bgcolor = $tt2_bgcolor;
	    }
	}

	print "<tr>\n",
	      "<td align=\"left\" valign=\"top\" bgcolor=\"${bgcolor}\">",
	      "<font size=\"${font_size}\" face=\"${font_face}\">";
	if ($u_artist ne '') {
	    print a({ -href => "${cddb_search_url}?f=artist&q=$u_artist",
		      -target => "_blank", -style => "text-decoration:none" },
		    escapeHTML($q_artist));
	}
	else {
	    print escapeHTML($q_artist);
	}
	print "</font>",
	      "</td>\n",
	      "<td align=\"left\" valign=\"top\" bgcolor=\"${bgcolor}\">",
	      "<font size=\"${font_size}\" face=\"${font_face}\">";
	if ($u_album ne '') {
	    print a({ -href => "${cddb_search_url}?f=disc&q=$u_album",
		      -target => "_blank", -style => "text-decoration:none" },
		    escapeHTML($q_album));
	}
	else {
	    print escapeHTML($q_album);
	}
	print "</font>",
	      "</td>\n",
	      "<td align=\"left\" valign=\"top\" bgcolor=\"${bgcolor}\">",
	      font({ -size => $font_size, -face => $font_face },
		   escapeHTML($q_genre)),
	      "</td>\n",
	      "<td align=\"center\" valign=\"top\" bgcolor=\"${bgcolor}\">",
	      font({ -size => $font_size, -face => $font_face }, $fields[7]),
	      "</td>\n",
	      "<td align=\"left\" valign=\"top\" bgcolor=\"${bgcolor}\">",
	      font({ -size => $font_size, -face => $font_face },
		   a({ -href => "${cddb_search_url}?f=track&q=$u_title",
		       -target => "_blank", -style => "text-decoration:none" },
		     escapeHTML($q_title))),
	      "</td>\n",
	      "</tr>\n";
    }

    print "</table>\n";
}


sub urlEncode
{
    my $tmpstr = shift;

    $tmpstr =~ s/(\W)/sprintf("%%%x", ord($1))/eg;
    return $tmpstr;
}


sub urlDecode
{
    my $tmpstr = shift;

    $tmpstr =~ tr/+?/ /;
    $tmpstr =~ s/%([0-9a-fA-F]{2})/pack('H2',$1)/ge;
    return $tmpstr;
}


sub stripQuestionMark
{
    my $tmpstr = shift;

    $tmpstr =~ s/\?//g;
    return $tmpstr;
}


sub nobr
{
    my $arg = '';
    my $tmpstr = '<nobr>';

    while ($arg = shift) {
	$tmpstr .= $arg;
    }
    $tmpstr .= '</nobr>';

    return $tmpstr;
}


